@@ -27,11 +27,14 @@ import java.nio.file.Paths;
 import java.security.MessageDigest;
 import java.security.Principal;
 import java.security.PublicKey;
+import java.security.SecureRandom;
 import java.security.cert.CertificateException;
-import java.util.List;
+import java.util.Arrays;
 import java.util.Collection;
+import java.util.List;
 import java.util.ArrayList;
 
+import org.apache.commons.codec.binary.Hex;
 import org.apache.commons.lang.StringUtils;
 import org.mozilla.jss.CryptoManager;
 import org.mozilla.jss.asn1.ANY;
@@ -47,15 +50,20 @@ import org.mozilla.jss.crypto.EncryptionAlgorithm;
 import org.mozilla.jss.crypto.InternalCertificate;
 import org.mozilla.jss.crypto.NoSuchItemOnTokenException;
 import org.mozilla.jss.crypto.ObjectNotFoundException;
+import org.mozilla.jss.crypto.PBEAlgorithm;
 import org.mozilla.jss.crypto.PrivateKey;
 import org.mozilla.jss.crypto.X509Certificate;
+import org.mozilla.jss.pkcs11.PK11Store;
 import org.mozilla.jss.pkcs12.AuthenticatedSafes;
 import org.mozilla.jss.pkcs12.CertBag;
 import org.mozilla.jss.pkcs12.PFX;
 import org.mozilla.jss.pkcs12.PasswordConverter;
 import org.mozilla.jss.pkcs12.SafeBag;
 import org.mozilla.jss.pkix.primitive.Attribute;
+import org.mozilla.jss.pkix.primitive.EncryptedPrivateKeyInfo;
 import org.mozilla.jss.util.Password;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
 
 import javax.naming.ldap.LdapName;
 import javax.naming.ldap.Rdn;
@@ -64,9 +72,89 @@ import org.mozilla.jss.netscape.security.x509.X509CertImpl;
 
 public class PKCS12Util {
 
+    private static Logger logger = LoggerFactory.getLogger(PKCS12Util.class);
 
+    public final static String NO_ENCRYPTION = "none";
+
+    public final static List<PBEAlgorithm> SUPPORTED_CERT_ENCRYPTIONS = Arrays.asList(new PBEAlgorithm[] {
+            null, // none
+            PBEAlgorithm.PBE_SHA1_RC2_40_CBC
+    });
+
+    public final static List<PBEAlgorithm> SUPPORTED_KEY_ENCRYPTIONS = Arrays.asList(new PBEAlgorithm[] {
+            PBEAlgorithm.PBE_PKCS5_PBES2,
+            PBEAlgorithm.PBE_SHA1_DES3_CBC
+    });
+
+    public final static PBEAlgorithm DEFAULT_CERT_ENCRYPTION = SUPPORTED_CERT_ENCRYPTIONS.get(0);
+    public final static String DEFAULT_CERT_ENCRYPTION_NAME = NO_ENCRYPTION;
+
+    public final static PBEAlgorithm DEFAULT_KEY_ENCRYPTION = SUPPORTED_KEY_ENCRYPTIONS.get(0);
+    public final static String DEFAULT_KEY_ENCRYPTION_NAME = DEFAULT_KEY_ENCRYPTION.toString();
+
+    SecureRandom random;
+    PBEAlgorithm certEncryption = DEFAULT_CERT_ENCRYPTION;
+    PBEAlgorithm keyEncryption = DEFAULT_KEY_ENCRYPTION;
     boolean trustFlagsEnabled = true;
 
+    public PKCS12Util() throws Exception {
+        random = SecureRandom.getInstance("pkcs11prng", "Mozilla-JSS");
+    }
+
+    public void setCertEncryption(String name) throws Exception {
+
+        for (PBEAlgorithm algorithm : SUPPORTED_CERT_ENCRYPTIONS) {
+
+            if (algorithm == null) {
+                if (NO_ENCRYPTION.equals(name)) {
+                    this.certEncryption = null;
+                    return;
+                }
+
+            } else if (algorithm.toString().equals(name)) {
+                this.certEncryption = algorithm;
+                return;
+            }
+        }
+
+        throw new Exception("Unsupported certificate encryption: " + name);
+    }
+
+    public void setCertEncryption(PBEAlgorithm algorithm) throws Exception {
+        this.certEncryption = algorithm;
+    }
+
+    public PBEAlgorithm getCertEncryption() {
+        return certEncryption;
+    }
+
+    public void setKeyEncryption(String name) throws Exception {
+
+        for (PBEAlgorithm algorithm : SUPPORTED_KEY_ENCRYPTIONS) {
+
+            if (algorithm == null) {
+                if (NO_ENCRYPTION.equals(name)) {
+                    this.keyEncryption = null;
+                    return;
+                }
+
+            } else if (algorithm.toString().equals(name)) {
+                this.keyEncryption = algorithm;
+                return;
+            }
+        }
+
+        throw new Exception("Unsupported key encryption: " + name);
+    }
+
+    public void setKeyEncryption(PBEAlgorithm algorithm) throws Exception {
+        this.keyEncryption = algorithm;
+    }
+
+    public PBEAlgorithm getKeyEncryption() {
+        return keyEncryption;
+    }
+
     public boolean isTrustFlagsEnabled() {
         return trustFlagsEnabled;
     }
@@ -102,7 +190,8 @@ public class PKCS12Util {
         icert.setObjectSigningTrust(PKCS12.decodeFlags(flags[2]));
     }
 
-    /** Add a private key to the PKCS #12 object.
+    /**
+     * Add a private key to the PKCS #12 object.
      *
      * The PKCS12KeyInfo object received comes about in two
      * different scenarios:
@@ -124,47 +213,93 @@ public class PKCS12Util {
     public void addKeyBag(PKCS12KeyInfo keyInfo, Password password,
             SEQUENCE encSafeContents) throws Exception {
 
+        byte[] keyID = keyInfo.getID();
+        logger.debug(" - Key ID: " + Hex.encodeHexString(keyID));
+
+        ASN1Value content;
+
         byte[] epkiBytes = keyInfo.getEncryptedPrivateKeyInfoBytes();
-        if (epkiBytes == null) {
-            PrivateKey k = keyInfo.getPrivateKey();
-            if (k == null) {
-                return;
+
+        if (epkiBytes != null) {
+            // private key already encrypted
+            content = new ANY(epkiBytes);
+
+        } else {
+            PrivateKey privateKey = keyInfo.getPrivateKey();
+            if (privateKey == null) {
+                throw new Exception("Missing private key for " + keyInfo.getFriendlyName());
             }
 
-            epkiBytes = CryptoManager.getInstance()
-                .getInternalKeyStorageToken()
-                .getCryptoStore()
-                .getEncryptedPrivateKeyInfo(
-                    /* For compatibility with OpenSSL and NSS >= 3.31,
-                     * do not BMPString-encode the passphrase when using
-                     * non-PKCS #12 PBE scheme such as PKCS #5 PBES2.
-                     *
-                     * The resulting PKCS #12 is not compatible with
-                     * NSS < 3.31.
-                     */
-                    null /* passConverter */,
-                    password,
-                    /* NSS has a bug that causes any AES CBC encryption
-                     * to use AES-256, but AlgorithmID contains chosen
-                     * alg.  To avoid mismatch, use AES_256_CBC. */
-                    EncryptionAlgorithm.AES_256_CBC,
-                    0 /* iterations (default) */,
-                    k);
+            CryptoToken token = CryptoManager.getInstance().getInternalKeyStorageToken();
+
+            if (keyEncryption == PBEAlgorithm.PBE_SHA1_DES3_CBC) {
+                content = create_EPKI_with_PBE_SHA1_DES3_CBC(token, privateKey, password);
+
+            } else if (keyEncryption == PBEAlgorithm.PBE_PKCS5_PBES2) {
+                content = create_EPKI_with_PBE_PKCS5_PBES2(token, privateKey, password);
+
+            } else {
+                throw new Exception("Unsupported key encryption: " + keyEncryption);
+            }
         }
 
         SET keyAttrs = createKeyBagAttrs(keyInfo);
 
-        SafeBag safeBag = new SafeBag(
-            SafeBag.PKCS8_SHROUDED_KEY_BAG, new ANY(epkiBytes), keyAttrs);
+        SafeBag safeBag = new SafeBag(SafeBag.PKCS8_SHROUDED_KEY_BAG, content, keyAttrs);
         encSafeContents.addElement(safeBag);
     }
 
+    public ASN1Value create_EPKI_with_PBE_SHA1_DES3_CBC(CryptoToken token, PrivateKey privateKey, Password password)
+            throws Exception {
+
+        // Use the same salt size and number of iterations as in pk12util.
+
+        byte[] salt = new byte[16];
+        random.nextBytes(salt);
+
+        return EncryptedPrivateKeyInfo.createPBE(
+                PBEAlgorithm.PBE_SHA1_DES3_CBC,
+                password,
+                salt,
+                100000, // iterations
+                new PasswordConverter(),
+                privateKey,
+                token);
+    }
+
+    public ASN1Value create_EPKI_with_PBE_PKCS5_PBES2(CryptoToken token, PrivateKey privateKey, Password password)
+            throws Exception {
+
+        CryptoStore store = token.getCryptoStore();
+
+        byte[] bytes = store.getEncryptedPrivateKeyInfo(
+                // For compatibility with OpenSSL and NSS >= 3.31,
+                // do not BMPString-encode the passphrase when using
+                // non-PKCS #12 PBE scheme such as PKCS #5 PBES2.
+                //
+                // The resulting PKCS #12 is not compatible with
+                // NSS < 3.31.
+                null, // password converter
+                password,
+                // NSS has a bug that causes any AES CBC encryption
+                // to use AES-256, but AlgorithmID contains chosen
+                // alg.  To avoid mismatch, use AES_256_CBC.
+                EncryptionAlgorithm.AES_256_CBC,
+                0, // iterations (default)
+                privateKey);
+
+        return new ANY(bytes);
+    }
+
     public void addCertBag(PKCS12CertInfo certInfo,
             SEQUENCE safeContents) throws Exception {
 
+        byte[] id = certInfo.getID();
+        logger.debug(" - Certificate ID: " + Hex.encodeHexString(id));
 
-        ASN1Value cert = new OCTET_STRING(certInfo.cert.getEncoded());
-        CertBag certBag = new CertBag(CertBag.X509_CERT_TYPE, cert);
+        X509CertImpl cert = certInfo.getCert();
+        ASN1Value certAsn1 = new OCTET_STRING(cert.getEncoded());
+        CertBag certBag = new CertBag(CertBag.X509_CERT_TYPE, certAsn1);
 
         SET certAttrs = createCertBagAttrs(certInfo);
 
@@ -188,20 +323,24 @@ public class PKCS12Util {
 
         SET attrs = new SET();
 
+        String friendlyName = keyInfo.getFriendlyName();
+        logger.debug("   Friendly name: " + friendlyName);
+
         SEQUENCE subjectAttr = new SEQUENCE();
         subjectAttr.addElement(SafeBag.FRIENDLY_NAME);
 
         SET subjectSet = new SET();
-        subjectSet.addElement(new BMPString(keyInfo.subjectDN));
+        subjectSet.addElement(new BMPString(friendlyName));
         subjectAttr.addElement(subjectSet);
 
         attrs.addElement(subjectAttr);
 
+        byte[] keyID = keyInfo.getID();
         SEQUENCE localKeyAttr = new SEQUENCE();
         localKeyAttr.addElement(SafeBag.LOCAL_KEY_ID);
 
         SET localKeySet = new SET();
-        localKeySet.addElement(new OCTET_STRING(keyInfo.id.toByteArray()));
+        localKeySet.addElement(new OCTET_STRING(keyID));
         localKeyAttr.addElement(localKeySet);
 
         attrs.addElement(localKeyAttr);
@@ -213,37 +352,46 @@ public class PKCS12Util {
 
         SET attrs = new SET();
 
+        String friendlyName = certInfo.getFriendlyName();
+        logger.debug("   Friendly name: " + friendlyName);
+
         SEQUENCE nicknameAttr = new SEQUENCE();
         nicknameAttr.addElement(SafeBag.FRIENDLY_NAME);
 
         SET nicknameSet = new SET();
-        nicknameSet.addElement(new BMPString(certInfo.nickname));
+        nicknameSet.addElement(new BMPString(friendlyName));
         nicknameAttr.addElement(nicknameSet);
 
         attrs.addElement(nicknameAttr);
 
-        if (certInfo.getID() != null) {
-            SEQUENCE localKeyAttr = new SEQUENCE();
-            localKeyAttr.addElement(SafeBag.LOCAL_KEY_ID);
+        String trustFlags = certInfo.getTrustFlags();
+        if (trustFlags != null && trustFlagsEnabled) {
+            logger.debug("   Trust flags: " + trustFlags);
 
-            SET localKeySet = new SET();
-            localKeySet.addElement(new OCTET_STRING(certInfo.id.toByteArray()));
-            localKeyAttr.addElement(localKeySet);
-
-            attrs.addElement(localKeyAttr);
-        }
-
-        if (certInfo.trustFlags != null && trustFlagsEnabled) {
             SEQUENCE trustFlagsAttr = new SEQUENCE();
             trustFlagsAttr.addElement(PKCS12.CERT_TRUST_FLAGS_OID);
 
             SET trustFlagsSet = new SET();
-            trustFlagsSet.addElement(new BMPString(certInfo.trustFlags));
+            trustFlagsSet.addElement(new BMPString(trustFlags));
             trustFlagsAttr.addElement(trustFlagsSet);
 
             attrs.addElement(trustFlagsAttr);
         }
 
+        byte[] keyID = certInfo.getKeyID();
+        if (keyID != null) {
+            logger.debug("   Key ID: " + Hex.encodeHexString(keyID));
+
+            SEQUENCE localKeyAttr = new SEQUENCE();
+            localKeyAttr.addElement(SafeBag.LOCAL_KEY_ID);
+
+            SET localKeySet = new SET();
+            localKeySet.addElement(new OCTET_STRING(keyID));
+            localKeyAttr.addElement(localKeySet);
+
+            attrs.addElement(localKeyAttr);
+        }
+
         return attrs;
     }
 
@@ -253,6 +401,7 @@ public class PKCS12Util {
 
     public void loadFromNSS(PKCS12 pkcs12, boolean includeKey, boolean includeChain) throws Exception {
 
+        logger.info("Loading certificates and keys from NSS database");
 
         CryptoManager cm = CryptoManager.getInstance();
         CryptoToken token = cm.getInternalKeyStorageToken();
@@ -263,94 +412,206 @@ public class PKCS12Util {
         }
     }
 
-    public void loadCertFromNSS(PKCS12 pkcs12, String nickname, boolean includeKey, boolean includeChain) throws Exception {
+    public void loadCertFromNSS(
+            PKCS12 pkcs12,
+            String nickname,
+            boolean includeKey,
+            boolean includeChain) throws Exception {
+
+        loadCertFromNSS(pkcs12, nickname, includeKey, includeChain, null);
+    }
+
+    public void loadCertFromNSS(
+            PKCS12 pkcs12,
+            String nickname,
+            boolean includeKey,
+            boolean includeChain,
+            String friendlyName) throws Exception {
 
         CryptoManager cm = CryptoManager.getInstance();
 
         X509Certificate[] certs = cm.findCertsByNickname(nickname);
+
+        if (certs == null || certs.length == 0) {
+            throw new Exception("Certificate not found: " + nickname);
+        }
+
         for (X509Certificate cert : certs) {
-            loadCertFromNSS(pkcs12, cert, includeKey, includeChain);
+            loadCertFromNSS(pkcs12, cert, includeKey, includeChain, friendlyName);
         }
     }
 
-    public void loadCertFromNSS(PKCS12 pkcs12, X509Certificate cert, boolean includeKey, boolean includeChain) throws Exception {
+    public void loadCertFromNSS(
+            PKCS12 pkcs12,
+            X509Certificate cert,
+            boolean includeKey,
+            boolean includeChain) throws Exception {
+
+        loadCertFromNSS(pkcs12, cert, includeKey, includeChain, null);
+    }
+
+    public void loadCertFromNSS(
+            PKCS12 pkcs12,
+            X509Certificate cert,
+            boolean includeKey,
+            boolean includeChain,
+            String friendlyName) throws Exception {
 
         CryptoManager cm = CryptoManager.getInstance();
 
-        BigInteger id = createLocalID(cert);
+        PKCS12CertInfo certInfo = createCertInfoFromNSS(cert, friendlyName);
+        pkcs12.addCertInfo(certInfo, true);
 
-        // load cert info
-        loadCertInfoFromNSS(pkcs12, cert, id, true);
+        byte[] id = certInfo.getID();
+        logger.debug(" - Certificate ID: " + Hex.encodeHexString(id));
+        logger.debug("   Friendly name: " + certInfo.getFriendlyName());
+        logger.debug("   Trust flags: " + certInfo.getTrustFlags());
 
         if (includeKey) {
             // load key info if exists
-            loadKeyInfoFromNSS(pkcs12, cert, id);
+
+            try {
+                PrivateKey privateKey = cm.findPrivKeyByCert(cert);
+
+                PKCS12KeyInfo keyInfo = createKeyInfoFromNSS(cert, privateKey, friendlyName);
+                pkcs12.addKeyInfo(keyInfo);
+
+                byte[] keyID = keyInfo.getID();
+                certInfo.setKeyID(keyID);
+                logger.debug("   Key ID: " + Hex.encodeHexString(keyID));
+
+            } catch (ObjectNotFoundException e) {
+                logger.debug("Certificate has no private key");
+            }
         }
 
         if (includeChain) {
             // load cert chain
             X509Certificate[] certChain = cm.buildCertificateChain(cert);
+            if (certChain.length > 1) {
+                logger.debug("   Certificate Chain:");
+            }
             for (int i = 1; i < certChain.length; i++) {
-                X509Certificate c = certChain[i];
-                BigInteger cid = createLocalID(c);
-                loadCertInfoFromNSS(pkcs12, c, cid, false);
+                X509Certificate caCert = certChain[i];
+
+                PKCS12CertInfo caCertInfo = createCertInfoFromNSS(caCert);
+                pkcs12.addCertInfo(caCertInfo, false);
+
+                byte[] caCertID = caCertInfo.getID();
+                logger.debug("   - Certificate ID: " + Hex.encodeHexString(caCertID));
+                logger.debug("     Friendly name: " + caCertInfo.getFriendlyName());
+                logger.debug("     Trust flags: " + caCertInfo.getTrustFlags());
             }
         }
     }
 
-    public void loadCertInfoFromNSS(PKCS12 pkcs12, X509Certificate cert, BigInteger id, boolean replace) throws Exception {
+    public PKCS12CertInfo createCertInfoFromNSS(
+            X509Certificate cert) throws Exception {
 
-        String nickname = cert.getNickname();
-
-        PKCS12CertInfo certInfo = new PKCS12CertInfo();
-        certInfo.id = id;
-        certInfo.nickname = nickname;
-        certInfo.cert = new X509CertImpl(cert.getEncoded());
-        certInfo.trustFlags = getTrustFlags(cert);
-
-        pkcs12.addCertInfo(certInfo, replace);
+        return createCertInfoFromNSS(cert, null);
     }
 
-    public void loadKeyInfoFromNSS(PKCS12 pkcs12, X509Certificate cert, BigInteger id) throws Exception {
+    public PKCS12CertInfo createCertInfoFromNSS(
+            X509Certificate cert,
+            String friendlyName) throws Exception {
 
-        String nickname = cert.getNickname();
+        // generate cert ID from SHA-1 hash of cert data
+        byte[] id = SafeBag.getLocalKeyIDFromCert(cert.getEncoded());
 
-        CryptoManager cm = CryptoManager.getInstance();
-
-        try {
-            PrivateKey privateKey = cm.findPrivKeyByCert(cert);
-
-            PKCS12KeyInfo keyInfo = new PKCS12KeyInfo(privateKey);
-            keyInfo.id = id;
-            keyInfo.subjectDN = cert.getSubjectDN().toString();
-
-            pkcs12.addKeyInfo(keyInfo);
-
-        } catch (ObjectNotFoundException e) {
+        if (friendlyName == null) {
+            friendlyName = cert.getNickname();
         }
+
+        X509CertImpl certImpl = new X509CertImpl(cert.getEncoded());
+        String trustFlags = getTrustFlags(cert);
+
+        PKCS12CertInfo certInfo = new PKCS12CertInfo();
+        certInfo.setID(id);
+        certInfo.setFriendlyName(friendlyName);
+        certInfo.setCert(certImpl);
+        certInfo.setTrustFlags(trustFlags);
+
+        return certInfo;
+    }
+
+    public PKCS12KeyInfo createKeyInfoFromNSS(
+            X509Certificate cert,
+            PrivateKey privateKey) throws Exception {
+
+        return createKeyInfoFromNSS(cert, privateKey, null);
+    }
+
+    public PKCS12KeyInfo createKeyInfoFromNSS(
+            X509Certificate cert,
+            PrivateKey privateKey,
+            String friendlyName) throws Exception {
+
+        byte[] keyID = privateKey.getUniqueID();
+
+        if (friendlyName == null) {
+            friendlyName = cert.getNickname();
+        }
+
+        PKCS12KeyInfo keyInfo = new PKCS12KeyInfo(privateKey);
+        keyInfo.setID(keyID);
+        keyInfo.setFriendlyName(friendlyName);
+
+        return keyInfo;
     }
 
     public PFX generatePFX(PKCS12 pkcs12, Password password) throws Exception {
 
-
-        SEQUENCE safeContents = new SEQUENCE();
-
-        for (PKCS12CertInfo certInfo : pkcs12.getCertInfos()) {
-            addCertBag(certInfo, safeContents);
-        }
-
-        SEQUENCE encSafeContents = new SEQUENCE();
-
-        for (PKCS12KeyInfo keyInfo : pkcs12.getKeyInfos()) {
-            addKeyBag(keyInfo, password, encSafeContents);
-        }
+        logger.info("Generating PKCS #12 data");
 
         AuthenticatedSafes authSafes = new AuthenticatedSafes();
-        authSafes.addSafeContents(safeContents);
-        authSafes.addSafeContents(encSafeContents);
+
+        Collection<PKCS12KeyInfo> keyInfos = pkcs12.getKeyInfos();
+        Collection<PKCS12CertInfo> certInfos = pkcs12.getCertInfos();
+
+        if (!keyInfos.isEmpty()) {
+            SEQUENCE keySafeContents = new SEQUENCE();
+
+            for (PKCS12KeyInfo keyInfo : keyInfos) {
+                addKeyBag(keyInfo, password, keySafeContents);
+            }
+
+            authSafes.addSafeContents(keySafeContents);
+        }
+
+        if (!certInfos.isEmpty()) {
+            SEQUENCE certSafeContents = new SEQUENCE();
+
+            for (PKCS12CertInfo certInfo : certInfos) {
+                addCertBag(certInfo, certSafeContents);
+            }
+
+            if (certEncryption == null) {
+                authSafes.addSafeContents(certSafeContents);
+
+            } else if (certEncryption == PBEAlgorithm.PBE_SHA1_RC2_40_CBC) {
+
+                byte[] salt = new byte[16];
+                random.nextBytes(salt);
+
+                authSafes.addEncryptedSafeContents(
+                        certEncryption,
+                        password,
+                        salt,
+                        100000, // iterations
+                        certSafeContents);
+
+            } else {
+                throw new Exception("Unsupported certificate encryption: " + certEncryption);
+            }
+        }
 
         PFX pfx = new PFX(authSafes);
-        pfx.computeMacData(password, null, 5);
+
+        // Use the same salt size and number of iterations as in pk12util.
+
+        byte[] salt = new byte[16];
+        random.nextBytes(salt);
+        pfx.computeMacData(password, salt, 100000);
 
         return pfx;
     }
@@ -359,6 +620,7 @@ public class PKCS12Util {
 
         PFX pfx = generatePFX(pkcs12, password);
 
+        logger.info("Storing PKCS #12 data into " + filename);
         ByteArrayOutputStream bos = new ByteArrayOutputStream();
         pfx.encode(bos);
         byte[] data = bos.toByteArray();
@@ -392,9 +654,10 @@ public class PKCS12Util {
                 ANY value = (ANY) values.elementAt(0);
 
                 ByteArrayInputStream bis = new ByteArrayInputStream(value.getEncoded());
-                BMPString subjectDN = (BMPString) new BMPString.Template().decode(bis);
+                BMPString friendlyName = (BMPString) new BMPString.Template().decode(bis);
 
-                keyInfo.subjectDN = subjectDN.toString();
+                keyInfo.setFriendlyName(friendlyName.toString());
+                logger.debug("   Friendly name: " + keyInfo.getFriendlyName());
 
             } else if (oid.equals(SafeBag.LOCAL_KEY_ID)) {
 
@@ -402,9 +665,13 @@ public class PKCS12Util {
                 ANY value = (ANY) values.elementAt(0);
 
                 ByteArrayInputStream bis = new ByteArrayInputStream(value.getEncoded());
-                OCTET_STRING keyID = (OCTET_STRING) new OCTET_STRING.Template().decode(bis);
+                OCTET_STRING keyIdAsn1 = (OCTET_STRING) new OCTET_STRING.Template().decode(bis);
 
-                keyInfo.id = new BigInteger(1, keyID.toByteArray());
+                byte[] keyID = keyIdAsn1.toByteArray();
+                keyInfo.setID(keyID);
+
+            } else {
+                logger.warn("   " + oid + ": " + attr.getValues());
             }
         }
 
@@ -420,8 +687,16 @@ public class PKCS12Util {
         OCTET_STRING certStr = (OCTET_STRING) certBag.getInterpretedCert();
         byte[] x509cert = certStr.toByteArray();
 
-        certInfo.cert = new X509CertImpl(x509cert);
-        Principal subjectDN = certInfo.cert.getSubjectDN();
+        // generate cert ID from SHA-1 hash of cert data
+        byte[] id = SafeBag.getLocalKeyIDFromCert(x509cert);
+        certInfo.setID(id);
+        logger.debug("   Certificate ID: " + Hex.encodeHexString(id));
+
+        X509CertImpl cert = new X509CertImpl(x509cert);
+        certInfo.setCert(cert);
+
+        Principal subjectDN = cert.getSubjectDN();
+        logger.debug("   Subject DN: " + subjectDN);
 
         SET bagAttrs = bag.getBagAttributes();
 
@@ -436,10 +711,10 @@ public class PKCS12Util {
                 ANY value = (ANY) values.elementAt(0);
 
                 ByteArrayInputStream bis = new ByteArrayInputStream(value.getEncoded());
-                BMPString nickname = (BMPString) (new BMPString.Template()).decode(bis);
-
-                certInfo.nickname = nickname.toString();
+                BMPString friendlyName = (BMPString) (new BMPString.Template()).decode(bis);
 
+                certInfo.setFriendlyName(friendlyName.toString());
+                logger.debug("   Friendly name: " + certInfo.getFriendlyName());
 
             } else if (oid.equals(SafeBag.LOCAL_KEY_ID)) {
 
@@ -447,9 +722,11 @@ public class PKCS12Util {
                 ANY value = (ANY) values.elementAt(0);
 
                 ByteArrayInputStream bis = new ByteArrayInputStream(value.getEncoded());
-                OCTET_STRING keyID = (OCTET_STRING) new OCTET_STRING.Template().decode(bis);
+                OCTET_STRING keyIdAsn1 = (OCTET_STRING) new OCTET_STRING.Template().decode(bis);
 
-                certInfo.id = new BigInteger(1, keyID.toByteArray());
+                byte[] keyID = keyIdAsn1.toByteArray();
+                certInfo.setKeyID(keyID);
+                logger.debug("   Key ID: " + Hex.encodeHexString(keyID));
 
             } else if (oid.equals(PKCS12.CERT_TRUST_FLAGS_OID) && trustFlagsEnabled) {
 
@@ -457,19 +734,20 @@ public class PKCS12Util {
                 ANY value = (ANY) values.elementAt(0);
 
                 ByteArrayInputStream is = new ByteArrayInputStream(value.getEncoded());
-                BMPString trustFlags = (BMPString) (new BMPString.Template()).decode(is);
+                BMPString trustFlagsAsn1 = (BMPString) (new BMPString.Template()).decode(is);
 
-                certInfo.trustFlags = trustFlags.toString();
+                String trustFlags = trustFlagsAsn1.toString();
+                certInfo.setTrustFlags(trustFlags);
+                logger.debug("   Trust flags: " + trustFlags);
+
+            } else {
+                logger.warn("   " + oid + ": " + attr.getValues());
             }
         }
 
-        if (certInfo.id == null) {
-            certInfo.id = createLocalID(x509cert);
-        }
-
-        if (certInfo.nickname == null) {
+        if (certInfo.getFriendlyName() == null) {
+            logger.debug("   Generating new friendly name");
             LdapName dn = new LdapName(subjectDN.getName());
-
             ArrayList<String> values = new ArrayList<String>();
 
             // The getRdns method returns the list in reverse order
@@ -481,7 +759,9 @@ public class PKCS12Util {
                 values.add(rdn.getValue().toString());
             }
 
-            certInfo.nickname = StringUtils.join(values, " - ");
+            String friendlyName = StringUtils.join(values, " - ");
+            certInfo.setFriendlyName(friendlyName);
+            logger.debug("   Friendly name: " + friendlyName);
         }
 
         return certInfo;
@@ -489,6 +769,7 @@ public class PKCS12Util {
 
     public void getKeyInfos(PKCS12 pkcs12, PFX pfx, Password password) throws Exception {
 
+        logger.debug("Load encrypted private keys:");
 
         AuthenticatedSafes safes = pfx.getAuthSafes();
 
@@ -503,6 +784,7 @@ public class PKCS12Util {
 
                 if (!oid.equals(SafeBag.PKCS8_SHROUDED_KEY_BAG)) continue;
 
+                logger.debug(" - Private key:");
                 PKCS12KeyInfo keyInfo = getKeyInfo(bag, password);
                 pkcs12.addKeyInfo(keyInfo);
             }
@@ -511,6 +793,7 @@ public class PKCS12Util {
 
     public void getCertInfos(PKCS12 pkcs12, PFX pfx, Password password) throws Exception {
 
+        logger.debug("Loading certificates:");
 
         AuthenticatedSafes safes = pfx.getAuthSafes();
 
@@ -525,6 +808,7 @@ public class PKCS12Util {
 
                 if (!oid.equals(SafeBag.CERT_BAG)) continue;
 
+                logger.debug(" - Certificate:");
                 PKCS12CertInfo certInfo = getCertInfo(bag);
                 pkcs12.addCertInfo(certInfo, true);
             }
@@ -533,6 +817,7 @@ public class PKCS12Util {
 
     public PKCS12 loadFromFile(String filename, Password password) throws Exception {
 
+        logger.info("Loading PKCS #12 file");
 
         Path path = Paths.get(filename);
         byte[] b = Files.readAllBytes(path);
@@ -574,7 +859,7 @@ public class PKCS12Util {
             throws CertificateException {
 
         for (PKCS12CertInfo certInfo : pkcs12.getCertInfos()) {
-            Principal certSubjectDN = certInfo.cert.getSubjectDN();
+            Principal certSubjectDN = certInfo.getCert().getSubjectDN();
 
             try {
                 LdapName certSubjdn  = new LdapName(certSubjectDN.toString());
@@ -595,22 +880,28 @@ public class PKCS12Util {
             String nickname,
             PKCS12KeyInfo keyInfo) throws Exception {
 
-
-        PKCS12CertInfo certInfo = pkcs12.getCertInfoByID(keyInfo.getID());
+        PKCS12CertInfo certInfo = pkcs12.getCertInfoByKeyID(keyInfo.getID());
         if (certInfo == null) {
+            logger.debug("Private key has no certificate, ignore");
             return;
         }
 
         CryptoManager cm = CryptoManager.getInstance();
         CryptoToken token = cm.getInternalKeyStorageToken();
-        CryptoStore store = token.getCryptoStore();
+        PK11Store store = (PK11Store)token.getCryptoStore();
 
-        X509Certificate cert = cm.importCACertPackage(certInfo.cert.getEncoded());
+        X509CertImpl certImpl = certInfo.getCert();
+        X509Certificate cert = cm.importCACertPackage(certImpl.getEncoded());
 
         // get public key
         PublicKey publicKey = cert.getPublicKey();
 
         byte[] epkiBytes = keyInfo.getEncryptedPrivateKeyInfoBytes();
+        if (epkiBytes == null) {
+            logger.debug(
+                "No EncryptedPrivateKeyInfo for key '"
+                + keyInfo.getFriendlyName() + "'; skipping key");
+        }
         try {
             // first true without BMPString-encoding the passphrase.
             store.importEncryptedPrivateKeyInfo(
@@ -627,7 +918,7 @@ public class PKCS12Util {
         // delete the cert again (it will be imported again later
         // with the correct nickname)
         try {
-            store.deleteCert(cert);
+            store.deleteCertOnly(cert);
         } catch (NoSuchItemOnTokenException e) {
             // this is OK
         }
@@ -645,33 +936,42 @@ public class PKCS12Util {
         CryptoToken ct = cm.getInternalKeyStorageToken();
         CryptoStore store = ct.getCryptoStore();
 
-        BigInteger id = certInfo.getID();
-        PKCS12KeyInfo keyInfo = pkcs12.getKeyInfoByID(id);
-
-        for (X509Certificate cert : cm.findCertsByNickname(certInfo.nickname)) {
+        String nickname = certInfo.getFriendlyName();
+        for (X509Certificate cert : cm.findCertsByNickname(nickname)) {
             if (!overwrite) {
                 return;
             }
             store.deleteCert(cert);
         }
 
+        X509CertImpl certImpl = certInfo.getCert();
         X509Certificate cert;
-        if (keyInfo != null) { // cert has key
-            importKey(pkcs12, password, certInfo.nickname, keyInfo);
 
-            cert = cm.importUserCACertPackage(certInfo.cert.getEncoded(), certInfo.nickname);
+        byte[] keyID = certInfo.getKeyID();
+
+        if (keyID != null) { // cert has key
+            logger.debug("Importing private key for " + certInfo.getFriendlyName());
+            PKCS12KeyInfo keyInfo = pkcs12.getKeyInfoByID(keyID);
+            importKey(pkcs12, password, certInfo.getFriendlyName(), keyInfo);
+
+            logger.debug("Importing user certificate " + certInfo.getFriendlyName());
+            cert = cm.importUserCACertPackage(
+                    certImpl.getEncoded(), certInfo.getFriendlyName());
 
         } else { // cert has no key
+            logger.debug("Importing CA certificate " + certInfo.getFriendlyName());
             // Note: JSS does not preserve CA certificate nickname
-            cert = cm.importCACertPackage(certInfo.cert.getEncoded());
+            cert = cm.importCACertPackage(certImpl.getEncoded());
         }
 
-        if (certInfo.trustFlags != null && trustFlagsEnabled)
-            setTrustFlags(cert, certInfo.trustFlags);
+        String trustFlags = certInfo.getTrustFlags();
+        if (trustFlags != null && trustFlagsEnabled) {
+            setTrustFlags(cert, trustFlags);
+        }
     }
 
     public void storeCertIntoNSS(PKCS12 pkcs12, Password password, String nickname, boolean overwrite) throws Exception {
-        Collection<PKCS12CertInfo> certInfos = pkcs12.getCertInfosByNickname(nickname);
+        Collection<PKCS12CertInfo> certInfos = pkcs12.getCertInfosByFriendlyName(nickname);
         for (PKCS12CertInfo certInfo : certInfos) {
             storeCertIntoNSS(pkcs12, password, certInfo, overwrite);
         }
@@ -681,6 +981,7 @@ public class PKCS12Util {
             PKCS12 pkcs12, Password password, boolean overwrite)
         throws Exception
     {
+        logger.info("Storing data into NSS database");
 
         for (PKCS12CertInfo certInfo : pkcs12.getCertInfos()) {
             storeCertIntoNSS(pkcs12, password, certInfo, overwrite);
